/*******************/
/* CIV .MAP READER */
/*   MAIN MODULE   */
/*******************/

#include <alloc.h>
#include <bios.h>
#include <dir.h>
#include <dos.h>
#include <fcntl.h>
#include <io.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys\stat.h>
#include "civ_de.h"

#define MAX_BYTES_PER_LINE	40
#define DECODING		0
#define ENCODING		1
#define IMAGE				0
#define TEXT				1
#define BITMAP			2

/* WINDOWS TYPEDEF'S */

typedef struct tagBITMAPFILEHEADER
{	WORD	bfType;
	DWORD	bfSize;
	WORD	bfReserved1;
	WORD	bfReserved2;
	DWORD	bfOffBits;
} BITMAPFILEHEADER;

typedef struct tagBITMAINFOHEADER
{	DWORD	biSize;
	DWORD	biWidth;
	DWORD	biHeight;
	WORD	biPlanes;
	WORD	biBitCount;
	DWORD	biCompression;
	DWORD	biSizeImage;
	DWORD	biXPelsPerMeter;
	DWORD	biYPelsPerMeter;
	DWORD	biClrUsed;
	DWORD	biClrImportant;
} BITMAPINFOHEADER;

typedef struct tagRGBQUAD
{	BYTE	rgbBlue;
	BYTE	rgbGreen;
	BYTE	rgbRed;
	BYTE	rgbReserved;
} RGBQUAD;

typedef struct tagBITMAPINFO
{	BITMAPINFOHEADER	bmiHeader;
	RGBQUAD						bmiColors[256];
} BITMAPINFO;

/* MESSAGES */

char	BannerMsg[]						=
	"\n\t\tͻ\n"
		"\t\t CIVILIZATION DECODER/ENCODER  Beta 1 \n"
		"\t\t       BoGaRT - Florence,  1993       \n"
		"\t\tͼ\n";

char	CantCreateMsg[]						= "Can't create %s\n";
char	DecodingMsg[]							= "Decoding %s into %s\n";
char	DoneMsg[]									= "\nDone\n";
char	NotDoneMsg[]							= "\nNot sucessful\n";
char	EncodingMsg[]							= "Encoding %s into %s\n";
char	FileNotFoundMsg[]					= "File %s not found\n";
char	FileExistsMsg[]						= "File %s exists; Overwrite ? (Y/N):";
char	IllegalArgMsg[]						= "Illegal argument %s\n";
char	IllegalWidthMsg[]					= "Illegal width %d\n";
char	NoEncodeTextMsg[]					= "Can't encode text files\n;";
char	NotBinaryMapMsg[]					= "%s is not a binary MAP file\n";
char	OriginalGridMsg[]					= "Grid width: %d, heigth: %d\n";
char	OutOfMemoryMsg[]					= "Not enough memory\n";
char	ReadErrorMsg[]						= "Error reading from input file\n";
char	ReorderingMsg[]						= "Reordering %s into %s\n";
char	UnsupportedBitmapMsg[]		= "Unsupported bitmap format\n";
char	WriteErrorMsg[]						= "Error writing to output file\n";
char	SyntaxMsg[]								=
		"Syntax: MAP fromFile toFile [/D|/E] [/T|/B|/I] [/Ppalettefile] [/Wnn]\n"
		" where  fromFile:  input file\n"
		"          toFile:  output file\n"
		"              /D:  Decode from .MAP  [optional: it is the default]\n"
		"              /E:  Encode to .MAP    [optional; default: decode from .MAP]\n"
		"   /Ppalettefile:  palette file name [optional]\n"
		"            /Wnn:  Grid width        [optional; no default]\n"
		" only for DECODING: decode to\n"
		"              /T:  Text file         [optional; default: binary image]\n"
		"              /B:  Windows .BMP file [optional; default: binary image]\n"
		"              /I:  Binary image      [optional; it is the default]\n\n";

/* STRINGS */

char	MAPFileExt[]				= ".MAP";
char	SVEFileExt[]				= ".SVE";
char	BinFileExt[]				= ".BIN";
char	BmpFileExt[]				= ".BMP";
char	TextFileExt[]				= ".LST";
char	NewLineString[]			= "\x0D\x0A";
char	NewPageString[1]		= "\x0C";

/* GLOBALS */

int	LineWidth = 0;
char	OutBuffer[MAXGRIDWIDTH];
char	InFileName[80];
char	AltFileName[80];
char	OutFileName[80];
CIVPALETTE1 AltPalette1;
CIVPALETTE3 AltPalette3;

/* PROTOTYPES */

int		BMPToPIC(int Handle, char far * CIVBuffer);
int		getkey(void);
char	GetYNKey(char * Message, char * Param);
void	LineOutput(int OutHandle, char far * CIVBuffer, int Length);
void	MAPOutput(int OutHandle, char far * CIVBuffer, int Format);
void	PICToBMP(int OutHandle, char far * CIVBuffer, unsigned GridX, unsigned GridY);

/*	***************
		*     MAIN    *
		***************/
void main(int argc, char * argv[])

{	int						ErrorLevel=0;
	struct ffblk	FFBlk;
	char *				InFileNamePntr;
	char *				OutFileNamePntr;
	char far *		CIVBuffer;
	int						File=0;
	int						i;
	int						InHandle	= -1;
	int						OutHandle = -1;
	int						OpCode    = DECODING;
	int						Format    = IMAGE;

	printf(BannerMsg);

	/* CHECK PARAMETERS */

	if(argc < 3 || argc > 7)
	{ printf(SyntaxMsg);
		exit(1);
	}

	for(i=1; i<argc; i++)
	{	if(argv[i][0] != '/')								/* if file name */
		{	if(!File)
				strcpy(InFileName, argv[i]);
			else
				strcpy(OutFileName, argv[i]);
			File++;
			continue;
		}
		switch(argv[i][1] & 0xDF)						/* ignore '/', param to upper case */
		{
		case 'D':
			OpCode = DECODING;
			break;
		case 'E':
			OpCode = ENCODING;
			break;
		case 'I':
			Format = IMAGE;
			break;
		case 'T':
			Format = TEXT;
			break;
		case 'B':
			Format = BITMAP;
			break;
		case 'P':
			strcpy(AltFileName, argv[i]+2);
			break;
		case 'W':
			LineWidth = atoi(argv[i]+2);
			if(LineWidth > MAXGRIDWIDTH || LineWidth < 1)
			{	printf(IllegalWidthMsg, LineWidth);
				exit(1);
			}
			break;
		default:
			printf(IllegalArgMsg, argv[i]);
			printf(SyntaxMsg);
			exit(1);
		}
	}
	if(File < 1 || File > 2)
	{ printf(SyntaxMsg);
		exit(1);
	}
	if(OpCode == ENCODING && Format == TEXT)
	{	printf(NoEncodeTextMsg);
		exit(1);
	}

	/* SET UP FILE NAMES */

	InFileNamePntr=strchr(InFileName, '.');

	if(OutFileName[0] == '\0')						/* if no output file name */
	{	strcpy(OutFileName, InFileName);		/* duplicate input file name */
		if( (OutFileNamePntr=strchr(OutFileName, '.')) != NULL)
		{	OutFileNamePntr[0] = '\0';				/* cut extension away */
			OutFileNamePntr = NULL;
		}
	}
	else
		OutFileNamePntr=strchr(OutFileName, '.');

	switch(OpCode)												/* provide extensions, if missing */
	{
	case ENCODING:
		if(InFileNamePntr == NULL)					/* no extension */
		{	switch(Format)
			{
			case IMAGE:
				strcpy(strchr(InFileName, '\0'), BinFileExt);
				break;
			case TEXT:
				strcpy(strchr(InFileName, '\0'), TextFileExt);
				break;
			case BITMAP:
				strcpy(strchr(InFileName, '\0'), BmpFileExt);
				break;
			}
		}
		if(OutFileNamePntr == NULL)
			strcpy(strchr(OutFileName, '\0'), MAPFileExt);
		break;

	case DECODING:
		if(InFileNamePntr == NULL)					/* no extension */
			strcpy(strchr(InFileName, '\0'), MAPFileExt);
		if(OutFileNamePntr == NULL)
		{	switch(Format)
			{
			case IMAGE:
				strcpy(strchr(OutFileName, '\0'), BinFileExt);
				break;
			case TEXT:
				strcpy(strchr(OutFileName, '\0'), TextFileExt);
				break;
			case BITMAP:
				strcpy(strchr(OutFileName, '\0'), BmpFileExt);
				break;
			}
		}
	}

	/* CHECK FOR INPUT FILE EXISTENCE */

	if(findfirst(InFileName, &FFBlk, 0))
	{	printf(FileNotFoundMsg, InFileName);
		exit(2);
	}

	/* CHECK FOR OUTPUT FILE EXISTENCE */

	if(!findfirst(OutFileName, &FFBlk, 0))
	{	if(GetYNKey(FileExistsMsg, OutFileName) == 'N')
			exit(2);
	}

	/* ALLOCATE BITMAP BUFFER */

	if( (CIVBuffer=farmalloc(CIVBUFFER_LENGTH)) == NULL)
	{	printf(OutOfMemoryMsg);
		exit(2);
	}

	/* LOAD ALTERNATE PALETTE, IF SPECIFIED */

	if(AltFileName[0])
	{	if(!ReadCIVFile(AltFileName, CIVBuffer, NULL))			/* don't get the bitmap */
		{	printf(FileNotFoundMsg, AltFileName);
			ErrorLevel = 2;
			goto Exit;
		}
		memcpy(&AltPalette1, &CIVPalette1, sizeof(CIVPALETTE1));
		memcpy(&AltPalette3, &CIVPalette3, sizeof(CIVPALETTE3));
	}
	else
	{	AltPalette1.Length = 0;
		AltPalette3.Length = 0;
	}

	/* READY FOR THE REAL JOB */

	switch(OpCode)
	{
	/************/
	/* DECODING */
	/************/

	case DECODING:
		printf(DecodingMsg, InFileName, OutFileName);

		/* DECODE */

		if(!ReadCIVFile(InFileName, CIVBuffer, CIV_GETBITMAP))
		{	printf(FileNotFoundMsg, InFileName);
			ErrorLevel = 2;
			goto Exit;
		}
		printf(OriginalGridMsg, GridWidth, GridHeight);

		/* RESTORE ALTERNATE PALETTE */

		if(AltPalette1.Length != 0)
			memcpy(&CIVPalette1, &AltPalette3, sizeof(CIVPALETTE1));
		if(AltPalette3.Length != 0)
			memcpy(&CIVPalette3, &AltPalette3, sizeof(CIVPALETTE3));

		/* OPEN OUTPUT FILE */

		if( (OutHandle=open(OutFileName, O_WRONLY | O_BINARY | O_CREAT | O_TRUNC,
						S_IREAD | S_IWRITE) ) == -1)
		{	printf(CantCreateMsg, OutFileName);
			ErrorLevel = 2;
			goto Exit;
		}

		/* SAVE */

		MAPOutput(OutHandle, CIVBuffer, Format);
		break;

	/************/
	/* ENCODING */
	/************/

	case ENCODING:
		printf(EncodingMsg, InFileName, OutFileName);
		InHandle = open(InFileName, O_RDONLY | O_BINARY);

		/* READ THE INPUT FILE AND ADJUST GRID DIMENSIONS */

		if(Format == BITMAP)
		{	if(!BMPToPIC(InHandle, CIVBuffer))
			{	ErrorLevel = 2;
				goto Exit;
			}
			if(LineWidth)											/* adjust grid dimensions */
			{	GridHeight = (unsigned)GridWidth*GridHeight / LineWidth;
				GridWidth  = LineWidth;
			}
		}
		else
		{	_dos_read(InHandle, CIVBuffer, CIVBUFFER_LENGTH, &(unsigned)i);
			if(LineWidth)											/* adjust grid dimensions */
			{	GridHeight = (unsigned)i / LineWidth;
				GridWidth  = LineWidth;
			}
			else
			{	GridHeight = (unsigned)i / 320;
				GridWidth  = 320;
			}
		}
		printf(OriginalGridMsg, GridWidth, GridHeight);

		/* RESTORE ALTERNATE PALETTES */

		if(AltPalette1.Length != 0)
			memcpy(&CIVPalette1, &AltPalette1, sizeof(CIVPALETTE1));
		else
			CIVPalette1.Length = 0;						/* invalidate default palette */
		if(AltPalette3.Length != 0)
			memcpy(&CIVPalette3, &AltPalette3, sizeof(CIVPALETTE3));
		else
			CIVPalette3.Length = 0;						/* invalidate default palette */

		/* ENCODE */

		WriteCIVFile(OutFileName, CIVBuffer);
		break;
	}

Exit:
	farfree(CIVBuffer);
	close(InHandle);
	close(OutHandle);
	if(!ErrorLevel)
		printf(DoneMsg);
	else
		printf(NotDoneMsg);
	exit(ErrorLevel);
}

/*	**************
		* MAP OUTPUT *
		************** */

void MAPOutput(int OutHandle, char far * CIVBuffer, int Format)
{	int					SliceWidth;
	int					Line;
	int					GridX;
	int					GridY;
	int					LineOffset;
	char far *	MAPPointer;
	int					RemainingWidth;

	/* IF THE USER SUPPLIED A LINE WIDTH, AJUST GRID DIMENSIONS */

	if(LineWidth)
	{	GridY = (unsigned)GridWidth*GridHeight / LineWidth;
		GridX  = LineWidth;
	}
	else
	{	GridX = GridWidth;
		GridY = GridHeight;
	}

	RemainingWidth = GridX;

	if(Format==BITMAP)
		PICToBMP(OutHandle, CIVBuffer, GridX, GridY);
	else
	{	for(LineOffset=0; LineOffset < GridX; LineOffset += SliceWidth)
		{	if(Format == TEXT)
				SliceWidth = (RemainingWidth >= MAX_BYTES_PER_LINE ? MAX_BYTES_PER_LINE : GridX);
			else
				SliceWidth = GridX;
			MAPPointer = CIVBuffer;
			for(Line=0; Line < GridY; Line++)
			{	movedata(FP_SEG(MAPPointer), FP_OFF(MAPPointer),
									FP_SEG(OutBuffer), FP_OFF(OutBuffer), GridX);
				if(Format == TEXT)
					LineOutput(OutHandle, MAPPointer+LineOffset, SliceWidth);
				else
					write(OutHandle, OutBuffer, GridX);
				MAPPointer += GridX;
			}
			RemainingWidth -= SliceWidth;
			if(Format==TEXT)						/* add page breaks between slices */
				write(OutHandle, NewPageString, sizeof(NewPageString));
		}
	}
}

/*	**************
		* PIC TO BMP *
		**************/

void	PICToBMP(int OutHandle, char far * CIVBuffer, unsigned GridX,
					unsigned GridY)
{	int						GridOutputWidth;
	int						i;
	BITMAPFILEHEADER	BitmapFileHeader;
	BITMAPINFO				BitmapInfo;

	/* FILL BITMAPFILEHEADER */

	BitmapFileHeader.bfType = 0x4D42;			/* "BM" */
	BitmapFileHeader.bfSize = 0;
	BitmapFileHeader.bfReserved1 = 0;
	BitmapFileHeader.bfReserved2 = 0;
	BitmapFileHeader.bfOffBits = sizeof(BITMAPFILEHEADER)+sizeof(BITMAPINFO);

	/* FILL BITMAPINFOHEADER */

	BitmapInfo.bmiHeader.biSize = 0x28;
	BitmapInfo.bmiHeader.biWidth = GridX;
	BitmapInfo.bmiHeader.biHeight = GridY;
	BitmapInfo.bmiHeader.biPlanes = 1;
	BitmapInfo.bmiHeader.biBitCount = 8;
	BitmapInfo.bmiHeader.biCompression =
	BitmapInfo.bmiHeader.biSizeImage =
	BitmapInfo.bmiHeader.biXPelsPerMeter =
	BitmapInfo.bmiHeader.biYPelsPerMeter = 0;
	BitmapInfo.bmiHeader.biClrUsed =
	BitmapInfo.bmiHeader.biClrImportant = 256;

	/* FILL PALETTE */

	for(i=CIVPalette3.FirstCol; i<=CIVPalette3.LastCol; i++)
	{	BitmapInfo.bmiColors[i].rgbRed  = CIVPalette3.Colors[i].Red  *4;
		BitmapInfo.bmiColors[i].rgbBlue = CIVPalette3.Colors[i].Blue *4;
		BitmapInfo.bmiColors[i].rgbGreen= CIVPalette3.Colors[i].Green*4;
		BitmapInfo.bmiColors[i].rgbReserved = 0;
	}

	write(OutHandle, &BitmapFileHeader, sizeof(BITMAPFILEHEADER));
	write(OutHandle, &BitmapInfo,       sizeof(BITMAPINFO));

	/* ROUND GRIDWIDTH TO UPPER LONG BOUNDARY */

	GridOutputWidth = ((GridX+3) >> 2) * 4;

	/* OUTPUT BITMAP FROM BOTTOM TO TOP LINE */

	for(i = GridY-1; i >= 0; i--)
	{	memset(OutBuffer, '\0', GridOutputWidth);
		movedata(FP_SEG(CIVBuffer+i*GridX), FP_OFF(CIVBuffer+i*GridX),
							FP_SEG(OutBuffer), FP_OFF(OutBuffer), GridX);
		if(write(OutHandle, OutBuffer, GridOutputWidth) != GridOutputWidth)
		{	printf(WriteErrorMsg);
			break;
		}
	}

	BitmapFileHeader.bfSize = tell(OutHandle);
	lseek(OutHandle, 0, SEEK_SET);
	write(OutHandle, &BitmapFileHeader, sizeof(BITMAPFILEHEADER));
}

/*	**************
		* BMP TO PIC *
		**************/

int	BMPToPIC(int Handle, char far * CIVBuffer)
{	int						GridInputWidth;
	int						i;
	BITMAPFILEHEADER	BitmapFileHeader;
	BITMAPINFO				BitmapInfo;

	/* GET BIMAP INFO */

	read(Handle, &BitmapFileHeader, sizeof(BITMAPFILEHEADER));
	read(Handle, &BitmapInfo,       sizeof(BITMAPINFO));

	if(	BitmapInfo.bmiHeader.biCompression != 0 ||
			BitmapInfo.bmiHeader.biBitCount != 8)
	{	printf(UnsupportedBitmapMsg);
		return(FALSE);
	}
	GridWidth = (int)BitmapInfo.bmiHeader.biWidth;
	GridHeight= (int)BitmapInfo.bmiHeader.biHeight;

	/* FILL PALETTE */

	CIVPalette3.Type     = 0x304D;
	CIVPalette3.Length   = 0x0302;
	CIVPalette3.FirstCol = 0;
	CIVPalette3.LastCol  = 255;
	for(i=CIVPalette3.FirstCol; i<=CIVPalette3.LastCol; i++)
	{	CIVPalette3.Colors[i].Red   = BitmapInfo.bmiColors[i].rgbRed   / 4;
		CIVPalette3.Colors[i].Blue  = BitmapInfo.bmiColors[i].rgbBlue  / 4;
		CIVPalette3.Colors[i].Green = BitmapInfo.bmiColors[i].rgbGreen / 4;
	}


	/* ROUND GRIDWIDTH TO UPPER LONG BOUNDARY */

	GridInputWidth = ((GridWidth+3) >> 2) * 4;

	/* INPUT BITMAP FROM BOTTOM TO TOP LINE */

	for(i = GridHeight-1; i >= 0; i--)
	{	if(read(Handle, OutBuffer, GridInputWidth) != GridInputWidth)
		{	printf(ReadErrorMsg);
			return(FALSE);
		}
		movedata(	FP_SEG(OutBuffer), FP_OFF(OutBuffer),
							FP_SEG(CIVBuffer+i*GridWidth), FP_OFF(CIVBuffer+i*GridWidth),
							GridWidth);
	}
	return(TRUE);
}

/*	***************
		* LINE OUTPUT *
		*************** */

void LineOutput(int OutHandle, char far * Buffer, int Length)
{	int		LinePtr;
	char	LineBuffer[150];
	char	* LineBufferPtr;

	for(LinePtr=0, LineBufferPtr=LineBuffer; LinePtr<Length; LinePtr++, LineBufferPtr+=3)
	{	sprintf(LineBufferPtr, "%02X ", *(Buffer+LinePtr));
		if(LineBufferPtr[0]=='0')
		{	LineBufferPtr[0] = '.';
			if(LineBufferPtr[1]=='0')
				LineBufferPtr[1] = '.';
		}
	}
	strcpy(LineBufferPtr-1, NewLineString);		/* overwrite last space */
	write(OutHandle, LineBuffer, strlen(LineBuffer));
}

/*	*******************
		* GET KEY PRESSED *
		*******************
Uses the BIOS to read the next keyboard character */

int getkey(void)
{	int key, lo, hi;

	key = bioskey(0);
	lo = key & 0X00FF;
	hi = (key & 0XFF00) >> 8;
	return((lo == 0) ? hi + 256 : lo);
}

/*	*******************
		*   GET 'Y'/'N'   *
		*******************
Outputs a message and read 'Y' or 'N' from the keyboard
return: 'Y' or 'N' */

char	GetYNKey(char * Message, char * Param)
{	char	Char;

	/* output interactive message to stderr !! */
	fprintf(stderr, Message, Param);
	while ( (Char=getkey() & 0x00DF) != 'Y' && Char != 'N');
	fprintf(stderr, " %c\n", Char);
	return(Char);
}
